:- module debug.
:- interface.
:- import_module game.
:- import_module io.

	% dump the game state to stderr, for debugging
:- pred output_state(game::in, io__state::di, io__state::uo) is det.

	% dump the board state to the current output stream,
	% given a way to dump board squares.
:- pred output_board(board(T)::in,
		(pred(x,y,T,io__state,io__state)::pred(in,in,in,di,uo) is det),
	io__state::di, io__state::uo) is det.

	% dump the state of a square (the element of an ordinary board)
:- pred output_square(game, x, y, square_kind, io__state, io__state).
:- mode output_square(in, in, in, in, di, uo) is det.

	% dump the state of a square in a distance influence map
:- pred output_distance(x, y, int, io__state, io__state).
:- mode output_distance(in, in, in, di, uo) is det.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module int, map, list, char, assoc_list, std_util.
:- import_module globals, options, getopt, bool.
:- import_module string, char.
:- import_module curs.
:- import_module output.

output_state(Game) -->
	globals__io_lookup_bool_option(dump_map, DumpMap),
	globals__io_lookup_bool_option(dump_packages, DumpPackages),
	globals__io_lookup_bool_option(dump_plan, DumpPlan),
	globals__io_lookup_bool_option(curses, Curses),
	io__stderr_stream(Stderr),
	io__set_output_stream(Stderr, Old),
	( { Curses = yes } ->
		curs__clear,
		curs__move(0,0)
	;
		[]
	),
	( { DumpMap = yes } ->
		output_board(Game)
	;
		[]
	),
	( { DumpPackages = yes } ->
		output_packages(Game^packages)
	;
		[]
	),
	( { DumpPlan = yes } ->
		output_plan(Game^plan)
	;
		[]
	),
	%output_enemies(Game^enemy_robots),
	%output_my_robot(Game^my_robot),
	io__set_output_stream(Old, _).

output_board(Game) -->
	output_board(Game^board, output_square(Game)).

output_board(Board, OutputSquare) -->
	{ y(Height) = Board^height },
	output_rows(y(Height + 1), Board, OutputSquare).

output_rows(y(Y), Board, OutputSquare) -->
	( { Y >= 0 } ->
		output_row(y(Y), Board, OutputSquare),
		output_rows(y(Y - 1), Board, OutputSquare)
	;
		[]
	).

output_row(Y, Board, OutputSquare) -->
	{ x(Width) = Board^width },
	output_cols(x(0), x(Width + 1), Y, Board, OutputSquare).

output_cols(x(X), x(MaxX), Y, Board, OutputSquare) -->
	( { X =< MaxX } ->
		{ Square = square(Board, x(X), Y) },
		OutputSquare(x(X), Y, Square),
		output_cols(x(X + 1), x(MaxX), Y, Board, OutputSquare)
	;
		output__nl
	).

	% A square can contain multiple things --
	% Multiple robots (at the start of the game different robots
	% can occupy the same square), packages, and perhaps a home base.
	% We can't display the full contents in a single character.
	% So we just pick the most important thing to say -- the first
	% in this list.  We also use uppercase/lowercase to distinguish between
	% things on a home base and things on a plain square.
:- type contents
	--->	me
	;	enemy
	;	package
	;	package_destination
	;	maybe_package
	;	none.

output_square(Game, X, Y, Square) -->
	{ Posn = (X,Y) },
	{ MyPosn = Game^my_robot^my_posn },
	{ EnemyPosns = Game^enemy_robot_posns },
	{ OnGroundPackagePosns = filter_map(package_on_ground,
		Game^packages^values) },
	{ PackageDestinationPosns = filter_map(package_destination,
		Game^packages^values) },
	{ DroppedByEnemyPackagePosns = filter_map(package_dropped_by_enemy,
		Game^packages^values) },
	{ Posn = MyPosn ->
		Contents = me
	; list__member(Posn, EnemyPosns) ->
		Contents = enemy
	; list__member(Posn, OnGroundPackagePosns) ->
		Contents = package
	; list__member(Posn, PackageDestinationPosns) ->
		Contents = package_destination
	; list__member(Posn, DroppedByEnemyPackagePosns) ->
		Contents = maybe_package
	;
		Contents = none
	},

	output_char(attr(Square, Contents), draw(Square, Contents)).

:- func package_on_ground(package_info) = coord is semidet.
package_on_ground(PackageInfo) = Posn :-
	PackageInfo ^ package_info_posn = on_the_ground(Posn).

:- func package_dropped_by_enemy(package_info) = coord is semidet.
package_dropped_by_enemy(PackageInfo) = Posn :-
	PackageInfo ^ package_info_posn = dropped_by_enemy(Posn).

:- func package_destination(package_info) = coord is semidet.
package_destination(PackageInfo) = Posn :-
	PackageInfo ^ package_info_dest = yes(Posn).

:- func draw(square_kind, contents) = character.
draw(lethal, _) 		= ('~').
draw(wall, _)			= ('#').
draw(home, me)			= ('R').
draw(home, enemy)		= ('E').
draw(home, package)		= ('P').
draw(home, package_destination)	= ('D').
draw(home, maybe_package) 	= ('M').
draw(home, none)		= ('@').
draw(plain, me)			= ('r').
draw(plain, enemy)		= ('e').
draw(plain, package)		= ('p').
draw(plain, package_destination)= ('d').
draw(plain, maybe_package)	= ('m').
draw(plain, none)		= ('.').

:- func attr(square_kind, contents) = attr.
attr(lethal, _) 		= fg_bg(blue, black).
attr(wall, _)			= normal.
attr(home, me)			= bold.
attr(home, enemy)		= fg_bg(red, black) `and` bold.
attr(home, package)		= fg_bg(yellow, black) `and` bold.
attr(home, package_destination)	= fg_bg(green, black).
attr(home, maybe_package) 	= normal.
attr(home, none)		= normal.
attr(plain, me)			= bold.
attr(plain, enemy)		= fg_bg(red, black) `and` bold.
attr(plain, package)		= fg_bg(yellow, black) `and` bold.
attr(plain, package_destination)= fg_bg(green, black).
attr(plain, maybe_package)	= normal.
attr(plain, none)		= normal.

output_distance(_X, _Y, Dist) -->
	{ Dist < 0 -> Char = ('-')
	; Dist > 9 -> Char = ('>')
	; char__int_to_digit(Dist, Char0) -> Char = Char0
	; Char = ('?')
	},
	output_char(Char).

output_plan(Plan) -->
	output_string("Plan: "),
	io__write_list(Plan, "", output_plan_step),
	output__nl.

:- pred output_plan_step(direction, io__state, io__state).
:- mode output_plan_step(in, di, uo) is det.
output_plan_step(east)  --> output_string("E ").
output_plan_step(west)  --> output_string("W ").
output_plan_step(north) --> output_string("N ").
output_plan_step(south) --> output_string("S ").

output_packages(Packages) -->
	output_string("Packages:\n"),
	io__write_list(map__values(Packages), "", output_package).

:- pred output_package(package_info, io__state, io__state).
:- mode output_package(in, di, uo) is det.
output_package(package(package_id(Id), Locn, MaybeWeight, MaybeDest)) -->
	{ ( MaybeWeight = yes(W) ->
		WeightMsg = string__int_to_string(W)
	;
		WeightMsg = "unknown"
	) },
	{ ( MaybeDest = yes((x(X),y(Y))) ->
		DestMsg = string__format("(%d,%d)",[i(X),i(Y)])
	;
		DestMsg = "unknown"
	) },
	{ PosnMsg = package_posn_to_string(Locn) },
	output_string(string__format("package %d, weight %s, dest %s, %s",
		[i(Id),s(WeightMsg),s(DestMsg),s(PosnMsg)])),
	output__nl.

package_posn_to_string(carried_by_me) = "carried by me".
package_posn_to_string(carried_by_enemy(robot_id(Id))) =
	string__format("carried by enemy (%d)", [i(Id)]).
package_posn_to_string(on_the_ground((x(X),y(Y)))) =
	string__format("on the ground at (%d,%d)", [i(X),i(Y)]).
package_posn_to_string(dropped_by_enemy((x(X),y(Y)))) =
	string__format("dropped or delivered by enemy, at (%d,%d)", [i(X),i(Y)]).
package_posn_to_string(unknown) = "unknown".

%-----------------------------------------------------------------------------%
