:- module game.
:- interface.
:- import_module list, int, std_util, map, random, array, influence_map.

%----------

% The complete state of the game

:- type game ---> game(
	board :: board,
	packages :: package_map,
	my_robot :: my_robot_info,
	enemy_robots :: enemy_robot_map,
	enemy_robot_posns :: list(posn),
	bases_visited :: bases_visited_set,
	base_distances :: base_distance_im,
	plan :: list(direction),
	planned_move :: maybe(direction),
	rand_seed :: random__supply
).

% Field access functions; these get/set individual packages
% or enemies within the package_map or the enemy_robot_map.
:- func enemy(robot_id, game) = enemy_robot_info.
:- func 'enemy :='(robot_id, game, enemy_robot_info) = game.

:- func package(package_id, game) = package_info.
:- func 'package :='(package_id, game, package_info) = game.

% Functions to construct and update the game state.
% These are called by main.m.

:- func init_game(board, config, server_response) = game.

:- pred update_game_state(server_response::in, game::in, game::out) is det.

:- pred update_packages(current_packages::in, game::in, game::out) is det.

%----------

% Coordinates

% Note: coordinates start at 1, not zero!
:- type coord ---> (x,y).
:- type posn == coord.
:- type x ---> x(int).
:- type y ---> y(int).

%----------

% The board

:- type board == board(square_kind).
:- type board(T).
:- inst uniq_board == uniq_array(uniq_array(ground)).
:- mode board_di == di(uniq_board).
:- mode board_uo == out(uniq_board).

:- func make_board(x, y, list(row(T)), T) = board(T).
:- func make_board(x, y, list(row)) = board.

:- func map_board(func(x, y, T1) = T2, board(T1)) = board(T2).

:- pred map_fold_board(pred(x, y, T1, T2, Acc, Acc), board(T1), board(T2),
		Acc, Acc).
:- mode map_fold_board(pred(in, in, in, out, in, out) is det,
		in, board_uo, in, out) is det.

	% note, this uses destructive update
:- func set_square(board(T), x, y, T) = board(T).
:- mode set_square(board_di, in, in, in) = board_uo is det.

:- type row(T) == list(T).
:- type row == row(square_kind).

:- func width(board(T)) = x.
:- func height(board(T)) = y.

:- pred valid_coord(board(T), coord).
:- mode valid_coord(in, in) is semidet.

:- pred valid_coord(board(T), x, y).
:- mode valid_coord(in, in, in) is semidet.

:- func square(board(T), coord) = T.
:- func square(board(T), x, y) = T.

	% These are the same as square except that if the corrdinates are
	% out of bounds then this returns the border value passed
	% to make_board/4 (e.g. wall, in the T=square_kind case).
	%
:- func safe_square(board(T), coord) = T.
:- func safe_square(board(T), x, y) = T.

:- type square_kind ---> plain ; lethal ; wall ; home.

%----------

% Packages

	% The package_map includes all known packages,
	% including those that are being carried by our robot
	% or by enemy robots.  There may be other packages
	% that we don't yet know about; these are not included in the map.
:- type package_map == map(package_id, package_info).

:- type package_id ---> package_id(int).

:- type package_info ---> package(
	package_info_id :: package_id,
	package_info_posn :: package_posn,
	% weight and destination, if known ("no" means unknown)
	package_info_weight :: maybe(weight),
	package_info_dest :: maybe(coord)		% "no" means unknown
).

:- type package_posn
	--->	carried_by_me
	;	carried_by_enemy(robot_id)
	;	on_the_ground(coord)
	;	dropped_by_enemy(coord)
			% Might be either on the ground,
			% or more likely was dropped onto destination
			% and has now disappeared.
			% XXX we should be smarter in figuring this out,
			% e.g. if we know the object's destination
			% of whether the enemy was pushed
	;	unknown.

:- type weight == int.

%----------

% Robots

:- type enemy_robot_map == map(robot_id, enemy_robot_info).

:- type robot_id ---> robot_id(int).

:- type my_robot_info ---> my_robot_info(
	my_id :: robot_id,
	my_posn :: coord,
	my_packages :: list(package_id),
	my_money :: money,
	my_capacity :: weight
	% my_history :: list(command)	% is it worth storing this?
).

:- type enemy_robot_info ---> enemy_robot_info(
	enemy_id :: robot_id,
	enemy_posn :: coord,
	enemy_packages :: list(package_id),
	enemy_history :: list(command), % XXX not yet set
	enemy_behaviour :: enemy_behaviour % XXX not yet set
).

:- type money == int.

%----------

	% We keep a set of the bases visited.
	% XXX This set is represented as a map.
	% We really want tree234_set
:- type bases_visited_set == map(coord, unit).

%----------

% Plans

	% Removes the remaining moves in the current plan, which will force
	% the strategy to form a new plan.
:- pred abandon_current_plan(game::in, game::out) is det.

	% Gets the next move, if there is currently a plan.  Otherwise
	% returns 'no'.
:- pred maybe_get_next_move(maybe(direction)::out, game::in, game::out) is det.

	% Maximum length of plans we try to find.
:- func plan_length_limit(game) = int.

%----------

% The server sends us a message giving the list of all the packages
% at our current location.

:- type current_packages == list(current_package).

:- type current_package --->
	current_package(
		cur_id :: package_id,
		cur_x :: x,
		cur_y :: y,
		cur_weight :: weight
	).

%----------

% Commands

:- type command ---> command(
	bid :: money,
	action :: action
).

:- type action
	--->	move(direction)
	;	pick(list(package_id))
	;	drop(list(package_id)).

	% nop_command = drop([]), i.e. it does nothing and moves nowhere.
:- func nop_command = command.

%----------

% Server responses

% The configuration consists of 
% 
%    a unique identifier, 
%    the maximum carrying capacity, 
%    the initial amount of money (used for bidding) 
%
% These are all integers and are sent on one line, separated by spaces, in the
% above order. 

:- type config --->
	config(
		config_my_id :: robot_id,
		config_my_capacity :: weight,
		config_my_money :: money
	).

:- type server_response == list(robot_update).

:- type robot_update --->
	robot_update(
		robot_update_id::robot_id,
		robot_update_actions::list(robot_action)
	).

:- type robot_action
	--->	moved(direction)
	;	picked(package_id)
	;	dropped(package_id)
	;	position(x,y)
	.

%----------

% Directions

:- type direction ---> north ; south ; east ; west.

:- func move(coord, direction) = coord.

:- func neighbours(coord) = list(coord).

:- func x_offset(direction) = int.
:- func y_offset(direction) = int.

%----------

% "behaviour": some kind of summary of the history,
% for use in predicting the robot's likely future moves.
% XXX not yet used
:- type enemy_behaviour ---> enemy_behaviour.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module array, exception, int, set.
:- import_module array_utils.

%-----------------------------------------------------------------------------%
% The complete state of the game
%-----------------------------------------------------------------------------%

enemy(Id, Game) =
	( EnemyInfo = Game ^ enemy_robots ^ elem(Id) ->
		EnemyInfo
	;
		% this is the first we've heard about this enemy;
		% it must have just appeared.
		enemy_robot_info(Id, (x(1),y(1)), [], [], enemy_behaviour)
	).
'enemy :='(Id, Game, EnemyInfo) =
	Game ^ enemy_robots ^ elem(Id) := EnemyInfo.

package(Id, Game) = 
	( PackageInfo = Game ^ packages ^ elem(Id) ->
		PackageInfo
	;
		% this is the first we've heard about this package
		package(Id, unknown, no, no)
	).
'package :='(Id, Game, PackageInfo) =
	Game ^ packages ^ elem(Id) := PackageInfo.

init_game(Board, Config, InitialServerResponse) = Game :-
	Config = config(MyId, MyCapacity, MyMoney),
	process_initial_server_response(MyId, InitialServerResponse,
		MyPosn, EnemyRobots),
	EnemyRobotPosns = list__map(func(E) = E^enemy_posn, EnemyRobots^values),
	% initially we start with no packages,
	% and we don't know where any of the other packages are
	MyPackages = [],
	Packages = map__init,
	% if we start on a base, it will be marked as visited when
	% we process the initial server response
	BasesVisited = map__init,
	BaseDistances = compute_base_distance_im(Board),
	MyRobot = my_robot_info(MyId, MyPosn, MyPackages, MyMoney, MyCapacity),
	random__init(42, Seed),
	Plan = [],
	PlannedMove = no,
	Game = game(Board, Packages, MyRobot, EnemyRobots, EnemyRobotPosns,
		BasesVisited, BaseDistances, Plan, PlannedMove, Seed).

%-----------------------------------------------------------------------------%

:- pred process_initial_server_response(robot_id, server_response, coord,
	map(robot_id, enemy_robot_info)).

process_initial_server_response(MyId, Updates, MyPosn, EnemyRobots) :-
	process_initial_server_response_2(MyId, Updates, no, MaybeMyPosn,
		map__init, EnemyRobots),
	( MaybeMyPosn = yes(MyPosn)
	; MaybeMyPosn = no, throw("no initial position")
	).

:- pred process_initial_server_response_2(robot_id, server_response,
		maybe(coord), maybe(coord), enemy_robot_map, enemy_robot_map).

process_initial_server_response_2(_MyId, [], MyPosn, MyPosn,
		EnemyRobots, EnemyRobots).
process_initial_server_response_2(MyId, [Update|Updates], MyPosn0, MyPosn,
		EnemyRobots0, EnemyRobots) :-
	Update = robot_update(Id, Actions),
	( Actions = [position(X,Y)] ->
		( Id = MyId ->
			MyPosn1 = yes((X,Y)),
			EnemyRobots1 = EnemyRobots0
		;
			EnemyId = Id,
			EnemyPosn = (X,Y),
			EnemyPackages = [],
			EnemyHistory = [],
			EnemyBehaviour = enemy_behaviour,
			EnemyRobot = enemy_robot_info(
				EnemyId, EnemyPosn, EnemyPackages,
				EnemyHistory, EnemyBehaviour),
			EnemyRobots1 = map__set(EnemyRobots0, Id, EnemyRobot),
			MyPosn1 = MyPosn0
		),
		process_initial_server_response_2(MyId, Updates, MyPosn1,
			MyPosn, EnemyRobots1, EnemyRobots)
	;
		throw("invalid initial server response")
	).

%-----------------------------------------------------------------------------%

update_game_state(Updates) -->
	list__foldl(apply_response, Updates),
	%
	% Delete dead robots.
	%
	% The spec says "A robot is listed if AND ONLY IF it was alive at
	% the beginning of the turn."  Hence any robots not listed must be
	% dead.
	%
	{ LiveRobotIds = list__map((func(U) = U^robot_update_id), Updates) },
	EnemyRobots =^ enemy_robots,
	^enemy_robots := map__select(EnemyRobots, list_to_set(LiveRobotIds)),
	Packages =^ packages,
	{ LivePackageIds = list__filter(
		package_is_live(Packages, LiveRobotIds),
		Packages ^ keys) },
	^packages := map__select(Packages, list_to_set(LivePackageIds)).

	% Fail if a package is being held by a dead robot.
	% Such packages disappear when the robot dies.
:- pred package_is_live(package_map::in, list(robot_id)::in, package_id::in)
	is semidet.
package_is_live(Packages, LiveRobots, PackageId) :-
	PackageInfo = Packages ^ det_elem(PackageId),
	\+ (
		PackageInfo ^ package_info_posn = carried_by_enemy(EnemyId),
		\+ (EnemyId `list__member` LiveRobots)
	).

:- pred apply_response(robot_update::in, game::in, game::out) is det.
apply_response(Update, GameState0, GameState) :-
	Update = robot_update(Id, Actions),
	MyId = GameState0^my_robot^my_id,
	( Id = MyId ->
		(
			% check if our reported actions match
			% with what we planned (if anything)
			GameState0^planned_move = yes(Direction),
			Actions \= [moved(Direction)]
		->
			% they don't -- we must have been pushed!
			abandon_current_plan(GameState0, GameState1)
		;
			GameState1 = GameState0
		),
		list__foldl(apply_my_action, Actions, GameState1,
			GameState)
	;
		list__foldl(apply_enemy_action(Id), Actions, GameState0,
			GameState)
	).

:- pred apply_enemy_action(robot_id::in, robot_action::in,
	game::in, game::out) is det.
apply_enemy_action(Id, position(X,Y)) -->
	^enemy(Id)^enemy_posn := (X,Y),
	recompute_enemy_robot_posns.
apply_enemy_action(Id, moved(Direction)) -->
	 OldPosn =^ enemy(Id)^enemy_posn,
	^enemy(Id)^enemy_posn := move(OldPosn, Direction),
	recompute_enemy_robot_posns.
apply_enemy_action(Id, picked(PackageId)) -->
	OldEnemyPackages =^ enemy(Id)^enemy_packages,
	^enemy(Id)^enemy_packages := [PackageId | OldEnemyPackages],
	^package(PackageId)^package_info_posn := carried_by_enemy(Id).
apply_enemy_action(Id, dropped(PackageId)) -->
	EnemyPackages =^ enemy(Id)^enemy_packages,
	EnemyPosn =^ enemy(Id)^enemy_posn,
	Packages =^ packages,
	PackageDest =^ package(PackageId)^package_info_dest,
	^enemy(Id)^enemy_packages := list__delete_all(EnemyPackages, PackageId),
	( { PackageDest = yes(EnemyPosn) } ->
		^packages := map__delete(Packages, PackageId)
	; { PackageDest = yes(_) } ->
		^package(PackageId)^package_info_posn :=
			on_the_ground(EnemyPosn)
	;
		% We don't know whether it was dropped at the destination or
		% not.  Probably it was, unless the enemy robot was
		% subject to pushing.  It's probably not worth trying for
		% these.
		^package(PackageId)^package_info_posn :=
			dropped_by_enemy(EnemyPosn)
	).

recompute_enemy_robot_posns -->
	Enemies =^ enemy_robots^values,
	^enemy_robot_posns := list__map(func(E) = E^enemy_posn, Enemies).

:- pred apply_my_action(robot_action::in, game::in, game::out) is det.
apply_my_action(position(X,Y)) -->
	{ NewPosn = (X,Y) },
	^my_robot^my_posn := NewPosn,
	maybe_mark_base_visited(NewPosn).
apply_my_action(moved(Direction)) -->
	OldPosn =^ my_robot^my_posn,
	{ NewPosn = move(OldPosn, Direction) },
	^my_robot^my_posn := NewPosn,
	maybe_mark_base_visited(NewPosn).
apply_my_action(picked(PackageId)) -->
	OldPackages =^ my_robot^my_packages,
	^my_robot^my_packages := [PackageId | OldPackages],
	^package(PackageId)^package_info_posn := carried_by_me.
apply_my_action(dropped(PackageId)) -->
	MyPackages =^ my_robot^my_packages,
	MyPosn =^ my_robot^my_posn,
	Packages =^ packages,
	PackageDest =^ package(PackageId)^package_info_dest,
	^my_robot^my_packages := list__delete_all(MyPackages, PackageId),
	( { PackageDest = yes(MyPosn) } ->
		^packages := map__delete(Packages, PackageId)
	; { PackageDest = yes(_) } ->
		^package(PackageId)^package_info_posn := on_the_ground(MyPosn)
	;
		% PackageDest = no (i.e. destination unknown).
		% This should never happen, since we should always have
		% seen the package destination before we picked it up.
		{ throw("dropped package with unknown destination") }
	).

:- pred maybe_mark_base_visited(posn, game, game).
maybe_mark_base_visited(Posn, Game0, Game) :-
	( square(Game0 ^ board, Posn) = home ->
		Game = (Game0 ^ bases_visited :=
			map__set(Game0 ^ bases_visited, Posn, unit))
	;
		Game = Game0
	).

%-----------------------------------------------------------------------------%

update_packages(Packages, GameState0, GameState) :-
	list__foldl(update_package, Packages, GameState0, GameState).

:- pred update_package(current_package::in, game::in, game::out) is det.
update_package(current_package(Id, X, Y, Weight)) -->
	{ Destination = (X,Y) },
	CurrentPosn =^ my_robot^my_posn,
	{ Package = package(Id, on_the_ground(CurrentPosn),
		yes(Weight), yes(Destination)) },
	^package(Id) := Package.

%-----------------------------------------------------------------------------%
% The board
%-----------------------------------------------------------------------------%

:- type board(T) == array(array(T)).

width(Board) = x(Board ^ elem(0) ^ size - 2).

height(Board) = y(Board ^ size - 2).

valid_coord(Board, x(X), y(Y)) :-
	X > 0,
	Y > 0,
	x(W) = width(Board),
	X =< W,
	y(H) = height(Board),
	Y =< H.

valid_coord(Board, (X, Y)) :-
	valid_coord(Board, X, Y).

square(Board, (X,Y)) = square(Board, X, Y).

square(Board, x(X), y(Y)) = Board^elem(Y)^elem(X).

set_square(Board0, x(X), y(Y), Sq) = Board :-
	Row0 = Board0^elem(Y),
	Row1 = array_inst_cast(Row0),
	Row = (Row1^elem(X) := Sq),
	Board = (Board0^elem(Y) := Row).

:- func array_inst_cast(array(T)) = array(T).
:- mode array_inst_cast(in) = array_uo is det.
:- pragma c_code(array_inst_cast(A0::in) = (A::array_uo),
	[will_not_call_mercury, thread_safe],
"
	A = A0;
").

safe_square(Board, (X, Y)) = safe_square(Board, X, Y).

safe_square(Board, X, Y) = square(Board, X, Y).

make_board(X, Y, Rows) = make_board(X, Y, Rows, wall).

make_board(x(X), _Y, Rows, Border) = array(
	[array(list__duplicate(X + 2, Border))] ++
	list__map((func(Row) = array([Border] ++ Row ++ [Border])), Rows) ++
	[array(list__duplicate(X + 2, Border))]
).

map_board(Func, Board) = 
	map_with_index(func(Y,Row) = map_with_index(
		func(X,Sq) = Func(x(X),y(Y),Sq), Row), Board).

map_fold_board(Pred, Board0, Board, BoardAcc0, BoardAcc) :-
	map_fold_with_index(MapRow, Board0, Board, BoardAcc0, BoardAcc),
	MapRow = (pred(Y::in, Row0::in, Row::out,
			RowAcc0::in, RowAcc::out) is det :-
		map_fold_with_index(MapSquare, Row0, Row, RowAcc0, RowAcc),
		MapSquare = (pred(X::in, Sq0::in, Sq::out,
				SqAcc0::in, SqAcc::out) is det :-
			Pred(x(X), y(Y), Sq0, Sq, SqAcc0, SqAcc))).

%-----------------------------------------------------------------------------%
% Plans
%-----------------------------------------------------------------------------%

abandon_current_plan -->
	^plan := [].

maybe_get_next_move(MaybeMove, GameState0, GameState) :-
	Plan0 = GameState0 ^ plan,
	(
		Plan0 = [],
		MaybeMove = no,
		GameState = GameState0
	;
		Plan0 = [Move | Plan],
		MaybeMove = yes(Move),
		GameState = GameState0 ^ plan := Plan
	).

	% XXX maybe we should vary this amount?
plan_length_limit(_) = 100.

%-----------------------------------------------------------------------------%
% Commands
%-----------------------------------------------------------------------------%

nop_command = command(1, drop([])).

%-----------------------------------------------------------------------------%
% Directions
%-----------------------------------------------------------------------------%

:- pragma inline(move/2).
move((x(X),y(Y)), Direction) =
	(x(X + x_offset(Direction)), y(Y + y_offset(Direction))).

:- pragma inline(neighbours/1).
neighbours(Posn) = [
	move(Posn, north),
	move(Posn, south),
	move(Posn, east),
	move(Posn, west)
].

:- pragma inline(x_offset/1).
x_offset(north) = 0.
x_offset(south) = 0.
x_offset(east) = 1.
x_offset(west) = -1.

:- pragma inline(y_offset/1).
y_offset(north) = 1.
y_offset(south) = -1.
y_offset(east) = 0.
y_offset(west) = 0.

%-----------------------------------------------------------------------------%
