%-----------------------------------------------------------------------------%
% This module implements the navigation strategy to be used by the robot.
%
% The strategy is based around an operator that takes a set of paths P
% from the current location and produces a new set of paths that can be
% reached from P by adding one extra step.
%
% For each path we collect certain information which can help us evaluate
% the path, e.g. the path length.  When two paths end at the same location,
% we prune the path which we evaluate to be the worst.
%
% If a step results in no new locations being reachable, then we have reached
% a fixpoint and we can select the best known path.  If we never remove any
% locations, then since there is a finite number of different locations on
% the board our algorithm must terminate.  However, we may wish to limit
% the number of iterations or otherwise prune the tree to improve performance.
%-----------------------------------------------------------------------------%

:- module navigate.
:- interface.
:- import_module game.
:- import_module list, set, std_util.

	% Values of this type represent paths starting from the current
	% position.  Note that the current position is implied; it is not
	% directly represented here.
	%
:- type path
	--->	path(
			value		:: float,
				% The value we determined for this path.

			path_info	:: path_info,
				% Accumulated information about the path.

			path_end	:: coord,
				% The final location.

			last_move	:: direction,
				% The final move before reaching the above
				% location.  Note carefully: we start the
				% iterations with a path that has the start
				% location as its path_end.  For this path,
				% the last_move field is meaningless.

			previous_moves	:: maybe(path)
				% If there were previous moves M then this
				% is yes(M), otherwise no.
		).

	% This type records the information that we accumulate for each
	% path.  We use this information to evaluate the path.
	%
:- type path_info
	--->	path_info(
			length		:: int,
				% The number of steps in the path, including
				% picking up and dropping packages.

			water		:: int,
				% The number of times we encounter water
				% next to us with open space directly
				% opposite.  These locations are dangerous.

			open		:: int,
				% The number of times we encounter open
				% space next to us with more open space
				% directly opposite.  At these locations,
				% we may be pushed and lose our packages.

			enemies		:: int,
				% The number of times we go near an enemy
				% robot.  This may be good or bad, depending
				% on our mood.

			packages	:: list(package_id),
				% The list of packages I would expect to be
				% carrying at the end of the path.

			delivered	:: list(package_id)
				% The list of packages I would expect to
				% have delivered by the end of the path.
		).

	% We evaluate paths in terms of what is in the path_info.  A
	% function of this type must be provided by the caller of the step
	% operator.  The reason this function is abstracted away is that
	% we may wish to use different strategies at different times.  E.g.:
	% 	- later in the game we may not consider the risk of water
	%	  to be as serious; 
	%	- when we hold valuable packages we may wish to avoid opponents
	%	  and open space.
	%	- we may wish to impose a maximum length if we don't have
	%	  much money left.
	%
:- type value_func == (func(path_info, coord) = float).

%-----------------------------------------------------------------------------%

	% Values of this type represent sets of paths starting from the
	% current location.  This is the domain which our operator works on.
	%
:- type path_set
	--->	path_set(
			frontier	:: list(path),
				% New paths that have just been reached.

			reached		:: set(coord),
				% The set of all locations we have reached,
				% either in the current frontier or some time
				% earlier.

			best		:: maybe({path, float})
				% The best path seen so far, and its valuation.
				% If we haven't taken any steps yet, this
				% will be 'no'.
		).

%-----------------------------------------------------------------------------%

	% find_best_path(Game, ValueFunc) returns the best path and value
	% available, given the following:
	%	Game		game data
	%	ValueFunc	evaluates paths.
	% The reason we return the value along with the path is that the
	% caller may wish to adjust its overall strategy based on how good
	% the best path is (e.g. if the result is no good it may wish to
	% be more aggressive).
	%
	% If there are no paths, return 'no'.
	%
:- func find_best_path(game, value_func) = maybe({path, float}).

	% Returns the list of moves in the order they should be performed.
	%
:- func get_moves(path) = list(direction).

:- pred path_length_is_greater(path::in, int::in) is semidet.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module int, float, exception, map.

find_best_path(Game, ValueFunc) = Best :-
	PathSet0 = init_path_set(Game ^ my_robot),
	Limit = max(Game^my_robot^my_money, plan_length_limit(Game)),
	PathSet = iterate_step_to_fixpoint(Game, ValueFunc, Limit, PathSet0),
	Best = PathSet ^ best.

get_moves(Path) = get_moves_2(Path, []).

:- func get_moves_2(path, list(direction)) = list(direction).

get_moves_2(Path, Moves0) = Moves :-
	MaybePreviousMoves = Path ^ previous_moves,
	(
		MaybePreviousMoves = yes(PreviousMoves),
		Moves = get_moves_2(PreviousMoves, [Path ^ last_move | Moves0])
	;
		MaybePreviousMoves = no,
		% Don't insert the last_move of this path, since it is a dummy.
		Moves = Moves0
	).

path_length_is_greater(Path, Length) :-
	(
		Length =< 0
	->
		true
	;
		MaybePrevious = Path ^ previous_moves,
		MaybePrevious = yes(Previous),
		path_length_is_greater(Previous, Length - 1)
	).

%-----------------------------------------------------------------------------%

:- func init_path_set(my_robot_info) = path_set.

init_path_set(MyInfo) = PathSet :-
	Pos = MyInfo ^ my_posn,
	PathInfo = init_path_info(MyInfo),
	Dummy = north,
	Frontier = [path(0.0, PathInfo, Pos, Dummy, no)],
	PathSet = path_set(Frontier, set__init, no).

:- func init_path_info(my_robot_info) = path_info.

init_path_info(MyInfo) = path_info(0, 0, 0, 0, MyInfo ^ my_packages, []).

%-----------------------------------------------------------------------------%

:- func iterate_step_to_fixpoint(game, value_func, int, path_set) = path_set.

iterate_step_to_fixpoint(Game, ValueFunc, Count, PathSet0) = PathSet :-
	(
		Count > 0
	->
		PathSet = iterate_to_fixpoint_2(Game, ValueFunc, Count,
			PathSet0)
	;
		PathSet = PathSet0
	).

:- func iterate_to_fixpoint_2(game, value_func, int, path_set) = path_set.

iterate_to_fixpoint_2(Game, ValueFunc, Count, PathSet0) = PathSet :-
	Frontier0 = PathSet0 ^ frontier,
	(
		Frontier0 = [],
		%
		% Reached a fixpoint, so return.
		%
		PathSet = PathSet0
	;
		Frontier0 = [_ | _],
		%
		% Calculate the new frontier.
		%
		Reached0 = PathSet0 ^ reached,
		Frontier = calculate_new_frontier(Game, ValueFunc, Reached0,
			Frontier0),

		%
		% Now determine the other fields of the path set.
		%
		Reached = add_locations(Reached0, Frontier),
		Best = acc_best_path(PathSet0 ^ best, Frontier),

		%
		% Keep iterating with the new path set.
		%
		PathSet1 = ((PathSet0 ^ best := Best)
				^ frontier := Frontier)
				^ reached := Reached,
		PathSet = iterate_step_to_fixpoint(Game, ValueFunc, Count - 1,
			PathSet1)
	).

%-----------------------------------------------------------------------------%

% When calculating the new frontier, we do the processing in two stages.
%
% The first stage is to calculate all of the new locations, and associate
% with each the path and direction that was used to get there.  Then we
% remove duplicate locations, and locations we have already been to.
%
% The second stage is to calculate new paths that lead to these new
% locations.  The paths have extra information, the path_infos, that we
% need to calculate for each new location.
%
% The reason we need to split this process into two stages is that we want
% to remove duplicates as soon as possible, before we calculate the
% path_infos.  If we don't do this, we will calculate a lot of path_infos
% unnecessarily.
%
% The intermediate_frontier type is used to hold the intermediate data
% produced by the first stage.  The second stage processes this data to
% produce a list of paths.

:- type intermediate_frontier == list(if_element).

:- type if_element
	--->	if_element(
			coord	:: coord,
				% End of the new path.

			dir	:: direction,
				% Direction followed after the base path.

			base	:: path
				% The base path.
		).

:- func calculate_new_frontier(game, value_func, set(coord), list(path))
	= list(path).

calculate_new_frontier(Game, ValueFunc, Reached, Frontier0) = Frontier :-
	%
	% stage 1
	%
	Intermediate0 = do_one_step(Game ^ board, Frontier0),
	list__sort(compare_if_elements, Intermediate0, Intermediate1),
	remove_adjacent_equivs(Intermediate1, Intermediate2),
	Intermediate = remove_reached_locations(Reached, Intermediate2),
	%
	% stage 2
	%
	RevFrontier = process_intermediate_frontier(Game, ValueFunc,
		Intermediate, []),
	Frontier = list__reverse(RevFrontier).

	% From each path in the given set take one step in each direction
	% possible.  If the resulting location is not on a wall or on water,
	% add the tuple of information to the output list.
	%
	% The output list is unsorted.
	%
:- func do_one_step(board, list(path)) = intermediate_frontier.

do_one_step(Board, Paths) = list__foldl(do_one_step_path(Board), Paths, []).

:- func do_one_step_path(board, path, intermediate_frontier)
	= intermediate_frontier.

do_one_step_path(Board, Path, Ps0) = Ps :-
	Ps1 = do_one_step_direction(Board, north, Path, Ps0),
	Ps2 = do_one_step_direction(Board, south, Path, Ps1),
	Ps3 = do_one_step_direction(Board, east, Path, Ps2),
	Ps = do_one_step_direction(Board, west, Path, Ps3).

:- func do_one_step_direction(board, direction, path, intermediate_frontier)
	= intermediate_frontier.

do_one_step_direction(Board, Direction, Path, Ps0) = Ps :-
	Pos = move(Path ^ path_end, Direction),
	(
		square_is_free(Board, Pos)
	->
		Ps = [if_element(Pos, Direction, Path) | Ps0]
	;
		Ps = Ps0
	).

	% square_is_free(Board, Pos) is true iff Pos is on the board and
	% the square kind there is plain or home.
:- pred square_is_free(board, coord).
:- mode square_is_free(in, in) is semidet.

square_is_free(Board, Pos) :-
	Kind = safe_square(Board, Pos),
	(
		Kind = plain
	;
		Kind = home
	).

:- pred compare_if_elements(if_element, if_element, comparison_result).
:- mode compare_if_elements(in, in, out) is det.

compare_if_elements(A, B, Result) :-
	compare(Result, A ^ coord, B ^ coord).

	% Removes equivalent elements of the list which are adjacent
	% to one another.  We keep the tuple with the cheapest path.
:- pred remove_adjacent_equivs(intermediate_frontier::in,
	intermediate_frontier::out) is det.

remove_adjacent_equivs([], []).
remove_adjacent_equivs([X | Xs], L) :-
	remove_adjacent_equivs_2(Xs, X, L).

:- pred remove_adjacent_equivs_2(intermediate_frontier::in, if_element::in,
	intermediate_frontier::out) is det.

remove_adjacent_equivs_2([], X, [X]).
remove_adjacent_equivs_2([X1 | Xs], X0, L) :-
	(
		compare_if_elements(X0, X1, (=))
	->
		%
		% Keep the most valuable path.
		%
		(
			X0^base^value > X1^base^value
		->
			remove_adjacent_equivs_2(Xs, X0, L)
		;
			remove_adjacent_equivs_2(Xs, X1, L)
		)
	;
		L = [X0 | L0],
		remove_adjacent_equivs_2(Xs, X1, L0)
	).

	% remove_reached_locations(Reached, Paths) is the list of paths
	% which are members of Paths and do not have their first element
	% in Reached.  The list order must be preserved.
:- func remove_reached_locations(set(coord), intermediate_frontier)
	= intermediate_frontier.

remove_reached_locations(Reached, Ps) = list__filter(NotReached, Ps) :-
	NotReached = (pred(A::in) is semidet :-
			\+ set__member(A^coord, Reached)
		).

	% Calculate path_infos for each member of the intermediate_frontier,
	% and build the new paths.  The output is in reverse order.
:- func process_intermediate_frontier(game, value_func, intermediate_frontier,
	list(path)) = list(path).

process_intermediate_frontier(_, _, [], Frontier) = Frontier.
process_intermediate_frontier(Game, ValueFunc, Intermediate, Frontier0)
		= Frontier :-

	Intermediate = [I | Intermediate0],
	PathInfo = update_path_info(Game, I^coord, I^base),
	Val = ValueFunc(PathInfo, I^coord),
	Path = path(Val, PathInfo, I^coord, I^dir, yes(I^base)),
	Frontier = process_intermediate_frontier(Game, ValueFunc, Intermediate0,
			[Path | Frontier0]).

:- func update_path_info(game, coord, path) = path_info.

update_path_info(Game, Pos, Path0) = PathInfo :-
	PathInfo0 = Path0 ^ path_info,
	PathInfo1 = PathInfo0 ^ length := PathInfo0 ^ length + 1,
	PathInfo2 = update_path_info_surrounds(Game, Pos, PathInfo1),
	PathInfo3 = update_path_info_home_bases(Game, Pos, PathInfo2),
	PathInfo = update_path_info_dests(Game, Pos, PathInfo3).

:- func update_path_info_surrounds(game, coord, path_info) = path_info.

update_path_info_surrounds(Game, Pos, PathInfo0) = PathInfo :-
	Board = Game ^ board,
	NP = move(Pos, north),
	SP = move(Pos, south),
	EP = move(Pos, east),
	WP = move(Pos, west),
	N = safe_square(Board, NP),
	S = safe_square(Board, SP),
	E = safe_square(Board, EP),
	W = safe_square(Board, WP),
	PathInfo1 = PathInfo0 ^ water := PathInfo0 ^ water
			+ count_water_open(N, S)
			+ count_water_open(S, N)
			+ count_water_open(E, W)
			+ count_water_open(W, E),
	PathInfo2 = PathInfo1 ^ open := PathInfo1 ^ open
			+ count_open_open(N, S)
			+ count_open_open(E, W),
	PathInfo = PathInfo2 ^ enemies := PathInfo2 ^ enemies
			+ count_enemies(Game, NP)
			+ count_enemies(Game, WP)
			+ count_enemies(Game, EP)
			+ count_enemies(Game, SP).

:- func count_water_open(square_kind, square_kind) = int.

count_water_open(Water, Open) = N :-
	(
		Water = lethal,
		is_open_space(Open)
	->
		N = 1
	;
		N = 0
	).

:- func count_open_open(square_kind, square_kind) = int.

count_open_open(K1, K2) = N :-
	(
		is_open_space(K1),
		is_open_space(K2)
	->
		N = 1
	;
		N = 0
	).

:- pred is_open_space(square_kind).
:- mode is_open_space(in) is semidet.

is_open_space(plain).
is_open_space(home).

:- func count_enemies(game, coord) = int.
count_enemies(Game, Posn) = EnemyCount :-
	list__foldl(
		(pred(EnemyPosn::in, Count0::in, Count::out) is det :-
			( EnemyPosn = Posn ->
				Count = Count0 + 1
			;
				Count = Count0
			)),
		Game^enemy_robot_posns, 0, EnemyCount).

:- func update_path_info_home_bases(game, coord, path_info) = path_info.

	% XXX need to do this properly.  If there are any known packages,
	% assume we can pick them up and add the ids to the list.
update_path_info_home_bases(_, _, Info) = Info.

:- func update_path_info_dests(game, coord, path_info) = path_info.

	% XXX need to do this properly.  If this square is a known destination
	% for a package we are carrying it, assume we drop it and add the
	% id to the delivered list.
update_path_info_dests(_, _, Info) = Info.

%-----------------------------------------------------------------------------%

	% Add the ends of the new paths to the set of reached locations.
:- func add_locations(set(coord), list(path)) = set(coord).

add_locations(Reached0, Frontier) = Reached :-
	Reached1 = set__sorted_list_to_set(list__map(get_path_end, Frontier)),
	Reached = set__union(Reached0, Reached1).

:- func get_path_end(path) = coord.

get_path_end(P) = P ^ path_end.

:- func acc_best_path(maybe({path, float}), list(path)) = maybe({path, float}).

acc_best_path(no, []) = no.
acc_best_path(no, [P | Ps]) = yes(BestVal) :-
	Value = P ^ value,
	BestVal = acc_best_path_2(P, Value, Ps).
acc_best_path(yes({Best0, Val0}), Ps) = yes(BestVal) :-
	BestVal = acc_best_path_2(Best0, Val0, Ps).

:- func acc_best_path_2(path, float, list(path)) = {path, float}.

acc_best_path_2(Best, Val, []) = {Best, Val}.
acc_best_path_2(Best0, Val0, [P | Ps]) = BestVal :-
	ValP = P ^ value,
	(
		ValP > Val0
	->
		BestVal = acc_best_path_2(P, ValP, Ps)
	;
		BestVal = acc_best_path_2(Best0, Val0, Ps)
	).

%-----------------------------------------------------------------------------%

