%-----------------------------------------------------------------------------%
:- module strategy_2.
:- interface.
:- import_module game.

%-----------------------------------------------------------------------------%

:- pred choose_command(current_packages::in, command::out,
	game::in, game::out) is det.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module list, int, float, random, exception, std_util, map.
:- import_module path.

	% XXX this is a really dumb algorithm... we should:
	%  - deliver any packages we hold for which this is the
	%    destination,
	%  - pickup packages that are here, and that we are able
	%    to carry,
	%  - decide whether we need to make a new plan, and if so then
	%    make a new plan,
	%  - look at what the current plan says we should do next,
	%  - decide on how much to bid.
	%
choose_command(CurrentPackages, Command) -->
	choose_action(CurrentPackages, Action),
	choose_bid(Action, Bid),
	{ Command = command(Bid, Action) }.

:- pred choose_action(current_packages::in, action::out,
	game::in, game::out) is det.

choose_action(CurrentPackages, Action, GameState0, GameState) :-
	(
		choose_drop(GameState0, Packages)
	->
		Action = drop(Packages),
		GameState = GameState0
	;
		choose_pick(CurrentPackages, GameState0, Packages)
	->
		Action = pick(Packages),
		GameState = GameState0
	;
		choose_move(Move, GameState0, GameState),
		Action = move(Move)
	).

%-----------------------------------------------------------------------------%

:- pred choose_drop(game::in, list(package_id)::out) is semidet.

choose_drop(Game, PackagesToDrop) :-
	CarriedIds = Game ^ my_robot ^ my_packages,
	list__filter(should_drop(Game), CarriedIds, PackagesToDrop),
	PackagesToDrop \= [].

:- pred should_drop(game::in, package_id::in) is semidet.
should_drop(Game, PackageId) :-
	MyPosn = Game ^ my_robot ^ my_posn,
	Game ^ package(PackageId) ^ package_info_dest = yes(MyPosn).

%-----------------------------------------------------------------------------%

:- pred choose_pick(current_packages::in, game::in, list(package_id)::out)
	is semidet.

choose_pick(CurrentPackages, GameState, PackagesToPickUp) :-
	% check current_packages for ones currently under us
	% that we can carry.
	list__filter_map(can_pick_up(GameState), CurrentPackages,
		PackagesToPickUp),
	PackagesToPickUp \= [].

:- pred can_pick_up(game::in, current_package::in, package_id::out) is semidet.
can_pick_up(Game, CurrentPackage, PackageId) :-
	Capacity = Game ^ my_robot ^ my_capacity,
	PackageId = CurrentPackage ^ cur_id,
	ThisWeight = package_weight(Game, PackageId), 
	CarriedIds = Game ^ my_robot ^ my_packages,
	Carrying = sum(list__map(package_weight(Game), CarriedIds)),
	Carrying + ThisWeight =< Capacity.

:- func sum(list(int)) = int.
sum(Ints) = list__foldl(func(X,Y)=X+Y, Ints, 0).

:- func package_weight(game, package_id) = weight is det.
package_weight(Game, PackageId) = Weight :-
	( Game^package(PackageId)^package_info_weight = yes(Weight0) ->
		Weight = Weight0
	;
		% This should never happen, since we should always find out
		% the package weights before we pick them up.
		% Actually it might happen if we are planning a path
		% which involves picking up an unknown package.
		% For those, just guess Weight = 1.
		% (XXX smarter guesses are possible...)
		Weight = 1
	).

%-----------------------------------------------------------------------------%

:- pred choose_move(direction::out, game::in, game::out) is det.

	% Decide if we need a new plan.  If so, create it. Then take
	% the first move of the current plan.
choose_move(Direction) -->
	maybe_get_next_move(MaybeMove),
	(
		{ MaybeMove = yes(Direction) }
	;
		{ MaybeMove = no },
		formulate_new_plan,
		% We avoid simply calling choose_move recursively, because
		% if we did and if for some reason we generate a new plan
		% that has no moves, we risk non-termination.
		maybe_get_next_move(MaybeMove2),
		(
			{ MaybeMove2 = yes(Direction) }
		;
			{ MaybeMove2 = no },
			random_direction(Direction)
		)
	),
	% record what we planned, so that we can abandon the plan
	% if we end up doing something different (because we got
	% pushed)
	^planned_move := yes(Direction).

	% Choose some goals, then generate paths and pick a path which
	% achieves these goals.  Later we may wish to assess whether
	% the chosen path is likely to achieve the goals, and if not,
	% pick some different goals.
:- pred formulate_new_plan(game::in, game::out) is det.

formulate_new_plan -->
	% choose_new_goals(Goals),
	=(GameState),
	% { NumEnemies = GameState^enemy_robots^count },
	{ Result = find_best_path(GameState ^ board, GameState^my_robot^my_posn,
		 (x(11), y(11)), unit_cost, 100.0) },
	{
		Result = search_aborted,
		% Urk, we are stuck.  Not much we can do.
		Moves = []
	;
		% XXX we don't check that this is a suitable path
		% (although it is the best we found).
		Result = found(_Cost, Moves)
	},
	^plan := Moves.

:- func unit_cost(coord) = float.
unit_cost(_) = 1.0.

:- pred random_direction(direction::out, game::in, game::out) is det.
random_direction(Direction) -->
	RS0 =^ rand_seed,
	{ random__random(R, RS0, RS) },
	^rand_seed := RS,
	{ I = R rem 4 },
	{ Direction0 = dir(I) ->
		Direction = Direction0
	;
		throw("dir failed")
	}.

:- func dir(int) = direction is semidet.
dir(0) = south.
dir(1) = north.
dir(2) = east.
dir(3) = west.

%-----------------------------------------------------------------------------%

:- pred choose_bid(action, int, game, game).
choose_bid(_Action, Bid, Game, Game) :-
	MyPosn = Game ^ my_robot ^ my_posn,
	MyNeighbours = neighbours(MyPosn),
	(
		% bid more if we are adjacent to other robots
		% (should we also do this if we are on top of other robots?)
		list__member(EnemyPosn, Game^enemy_robot_posns),
		list__member(EnemyPosn, MyNeighbours)
	->
		% bid even more if there are lethal squares next to us
		(
			list__member(Posn, MyNeighbours),
			square(Game^board, Posn) = lethal
		->
			Bid = 5
		;
			Bid = 2
		)
	;
		% bid negative if we are about to move into a square that
		% is adjacent to other robots
		% also bid (+/-)1 in the normal case
		Bid = -1
	).

%-----------------------------------------------------------------------------%
%
% Goals:
% - visit unvisited home square (to discover packages)
% - collect packages
% - deliver packages
% - push enemies
% - kill enemies
%
% - visit unvisited home square (to discover packages)
%	estimated value:
%		max(remaining capacity,
%		average of previous package values on home squares)
%	estimated cost:
%		average of previous distances from home square to dest
%		width of board
%		average distance from home square to other squares
%		
% - collect packages
%	but only if those packages are deliverable
%		(not if destination is unreachable,
%		or not reachable within our fuel limit)
%	estimated value: value of package
%	estimated cost: distance to package + 1 (to pick it up) +
%		distance to package destination + 1 (to drop it)
%
% - deliver packages
%	estimated cost: distance to distination + 1 (to drop it)
%	estimated value: value of package
%
% - push enemies
%
% - kill enemies
%
%-----------------------------------------------------------------------------%

:- type goals == list(goal).
:- type goal ---> goal. % XXX

:- pred choose_new_goals(goals::out, game::in, game::out) is det.

choose_new_goals(Goals) -->
	% XXX
	{ Goals = [] }.

%-----------------------------------------------------------------------------%
/*
:- func assess_path_against_goals(game, int, goals, path_info, coord) = float.

	% This is the heuristic we use to determine which of two paths best
	% achieves our goals.  For the economic goals (that is, picking up
	% and delivering packages) this function should, roughly speaking,
	% estimate the rate at which we would profit by following the given
	% path.  The rate of profit is estimated as follows:
	%	(value of deliveries - cost of travel) / number of moves
	% This is equivalent to:
	%	(value of deliveries / number of turns) - travel cost per turn
	% The reason we need the second term in this formula, the travel cost
	% per turn, is because different paths may cost different amounts.
	% E.g., a path that doesn't go near enemies or water is likely to be
	% cheaper, since we won't need to bid high to get out of tight spots.
	%
assess_path_against_goals(Game, NumEnemies, _, PathInfo, Posn) = Result :-
	% start with value for visiting unvisited home bases
	Sum0 = visit_base_value(Game, Posn),
	% add value for picking up and delivering packages
	PackageIds = map__keys(Game^packages),
	list__foldl(package_value(Posn, PathInfo, Game), PackageIds, Sum0, Sum),
	Cost = compute_cost(PathInfo, NumEnemies),
	% compute benefit/cost ratio
	Result = (Sum - Cost) / float(PathInfo ^ length).

:- func compute_cost(path_info, int) = float.
compute_cost(PathInfo, NumEnemies) =
	% XXX are these good values?
	float(PathInfo ^ water) * float(NumEnemies) +
	float(PathInfo ^ enemies) * 4.0 +
	float(PathInfo ^ open) * 0.1.

:- func visit_base_value(game, coord) = float.
visit_base_value(Game, Posn) = Value :-
	(
		square(Game^board, Posn) = home,
		% only value bases that we haven't visited before
		\+ map__search(Game^bases_visited, Posn, _)
	->
		% how much should we value unvisited home bases?
		Value = 10.0
	;
		Value = 0.0
	).

:- pred package_value(coord, path_info, game, package_id, float, float).
:- mode package_value(in, in, in, in, in, out) is det.
package_value(Posn, PathInfo, Game, PackageId, Sum0, Sum) :-
	map__lookup(Game^packages, PackageId, PackageInfo),
	PackageLocn = PackageInfo ^ package_info_posn,
	PackageDest = PackageInfo ^ package_info_dest,
	(
		PackageLocn = carried_by_me,
		PackageDest = yes(Posn),
		PackageInfo ^ package_info_weight = yes(Weight1)
	->
		Value = float(Weight1)
	;
		PackageLocn = on_the_ground(Posn),
		PackageInfo ^ package_info_weight = yes(Weight2),
		can_pickup_package(PathInfo ^ packages, Game, Weight2)
	->
		% probably it will take some extra work to deliver
		% the package, so only count 1/3 of its weight
		% for picking it up
		Value = float(Weight2) / 3.0
	;
		PackageLocn = dropped_by_enemy(Posn)
	->
		% these are unlikely to be worth investigating
		Value = 0.001
	;
		Value = 0.0
	),
	Sum = Sum0 + Value.

can_pickup_package(CurrentPackages, Game, PackageWeight) :-
	Capacity = Game ^ my_robot ^ my_capacity,
	CurrentWeight = sum_package_weights(Game, CurrentPackages),
	Capacity >= CurrentWeight + PackageWeight.

:- func sum_package_weights(game, list(package_id)) = int.
sum_package_weights(_Game, []) = 0.
sum_package_weights(Game, [Package|Packages]) =
	package_weight(Game, Package) + sum_package_weights(Game, Packages).
*/
